/*
** pork_buddy.c - Buddy list / permit / deny management.
** Copyright (C) 2002-2003 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <config.h>

#include <unistd.h>
#include <ncurses.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include <libfaim/aim.h>
#include <libfaim/faimconfig.h>

#include <pork_missing.h>
#include <pork_util.h>
#include <pork_list.h>
#include <pork_buddy.h>
#include <pork_set.h>
#include <pork_slist.h>
#include <pork_buddy_list.h>
#include <pork_acct.h>
#include <pork_cstr.h>
#include <pork_misc.h>
#include <pork_aim.h>
#include <pork_screen_cmd.h>

static int buddy_hash_compare(void *l, void *r) {
	char *str = l;
	struct buddy *buddy = r;

	return (strcasecmp(str, buddy->nname));
}

static void buddy_hash_cleanup(void *param __notused, void *data) {
	struct buddy *buddy = data;

	free(buddy->nname);
	free(buddy->name);
	free(buddy);
}

static void group_cleanup(void *param __notused, void *data) {
	struct bgroup *group = data;

	/*
	** The memory allocated for the buddies in the member_list of
	** each group is freed when the buddy hash is destroyed, so
	** we don't have to do it here.
	*/

	dlist_destroy(group->member_list, NULL, NULL);
	free(group->name);
	free(group);
}

static void generic_cleanup(void *param __notused, void *data) {
	free(data);
}

static int buddy_rights_compare(void *l, void *r) {
	char *str1 = l;
	char *str2 = r;

	return (strcasecmp(str1, str2));
}

static int group_compare(void *l, void *r) {
	char *name = l;
	struct bgroup *gr = r;

	return (strcasecmp(name, gr->name));
}

static int buddy_hash_remove(struct buddy_pref *pref, const char *screen_name) {
	char buddy[NUSER_LEN];
	u_int32_t hash;

	normalize(screen_name, buddy, sizeof(buddy));
	hash = string_hash(buddy, pref->buddy_hash.order);

	return (hash_remove(&pref->buddy_hash, buddy, hash));
}

int buddy_remove(struct aim_priv *priv, const char *screen_name, int aim) {
	struct buddy *buddy;
	dlist_t *list_node;
	char nname[NUSER_LEN];
	struct buddy_pref *pref = priv->buddy_pref;

	if (pref == NULL)
		return (-1);

	normalize(screen_name, nname, sizeof(nname));
	buddy = buddy_find(priv, nname);
	if (buddy == NULL)
		return (-1);

	list_node = dlist_find(buddy->group->member_list,
					buddy->nname, buddy_hash_compare);

	if (list_node == NULL)
		return (-1);

	if (aim)
		aim_buddy_remove(priv, buddy);

	buddy->group->member_list = dlist_remove(buddy->group->member_list,
									list_node);
	buddy->group->num_members--;
	blist_del(priv->blist, buddy);

	return (buddy_hash_remove(pref, buddy->name));
}

/*
** Add the buddy whose screen name is "screen_name" to the specified
** group.
*/

struct buddy *buddy_add(struct aim_priv *priv,
						const char *screen_name,
						struct bgroup *group,
						int aim)
{
	struct buddy *buddy;
	char nname[NUSER_LEN];
	struct buddy_pref *pref = priv->buddy_pref;

	if (pref == NULL)
		return (NULL);

	normalize(screen_name, nname, sizeof(nname));
	if (buddy_find(priv, nname) != NULL)
		return (NULL);

	buddy = xcalloc(1, sizeof(*buddy));
	buddy->name = xstrdup(screen_name);
	buddy->nname = xstrdup(nname);
	buddy->group = group;
	buddy->refnum = group->high_refnum++;

	group->num_members++;
	group->member_list = dlist_add_tail(group->member_list, buddy);

	hash_add(&pref->buddy_hash, buddy,
		string_hash(buddy->nname, pref->buddy_hash.order));

	if (aim)
		aim_buddy_add(priv, buddy);

	return (buddy);
}

struct bgroup *group_find(struct aim_priv *priv, char *group_name) {
	dlist_t *node;
	struct buddy_pref *pref = priv->buddy_pref;

	if (pref == NULL)
		return (NULL);

	node = dlist_find(pref->group_list, group_name, group_compare);
	if (node == NULL)
		return (NULL);

	return (node->data);
}

struct bgroup *group_add(struct aim_priv *priv, char *group_name) {
	struct bgroup *group;
	struct buddy_pref *pref = priv->buddy_pref;

	if (pref == NULL)
		return (NULL);

	group = group_find(priv, group_name);
	if (group != NULL)
		return (group);

	group = xcalloc(1, sizeof(*group));
	group->name = xstrdup(group_name);
	group->refnum = pref->next_group_refnum++;
	pref->group_list = dlist_add_tail(pref->group_list, group);

	blist_add_group(priv->blist, group);

	return (group);
}

int group_remove(struct aim_priv *priv, char *group_name, int aim) {
	struct bgroup *group;
	dlist_t *memb_list;
	dlist_t *node;
	struct buddy_pref *pref = priv->buddy_pref;

	if (pref == NULL)
		return (-1);

	node = dlist_find(pref->group_list, group_name, group_compare);
	if (node == NULL)
		return (-1);

	group = node->data;
	memb_list = group->member_list;

	while (memb_list != NULL) {
		dlist_t *next = memb_list->next;
		struct buddy *buddy = memb_list->data;

		buddy_remove(priv, buddy->name, aim);
		memb_list = next;
	}

	pref->group_list = dlist_remove(pref->group_list, node);
	blist_del_group(priv->blist, group);

	return (0);
}

struct buddy *buddy_find(struct aim_priv *priv, const char *screen_name) {
	dlist_t *node;
	char buddy[NUSER_LEN];
	u_int32_t hash;
	struct buddy_pref *pref = priv->buddy_pref;

	if (pref == NULL)
		return (NULL);

	normalize(screen_name, buddy, sizeof(buddy));
	hash = string_hash(buddy, pref->buddy_hash.order);

	node = hash_find(&pref->buddy_hash, buddy, hash);
	if (node != NULL)
		return (node->data);

	return (NULL);
}

void buddy_alias(	struct aim_priv *priv, 
					struct buddy *buddy,
					char *alias, 
					int aim)
{
	if (!strcmp(buddy->name, alias))
		return;

	free(buddy->name);
	buddy->name = xstrdup(alias);

	blist_update_label(priv->blist, buddy->blist_line);

	if (aim)
		aim_buddy_alias(priv, buddy);
}

void buddy_clear_block(struct aim_priv *priv) {
	struct buddy_pref *pref = priv->buddy_pref;

	dlist_destroy(pref->block_list, NULL, generic_cleanup);
	pref->block_list = NULL;
}

void buddy_clear_permit(struct aim_priv *priv) {
	struct buddy_pref *pref = priv->buddy_pref;

	dlist_destroy(pref->permit_list, NULL, generic_cleanup);
	pref->permit_list = NULL;
}

int buddy_init(struct aim_priv *priv) {
	struct buddy_pref *pref;

	if (priv->buddy_pref != NULL)
		return (-1);

	pref = xcalloc(1, sizeof(*pref));
	priv->buddy_pref = pref;

	hash_init(&pref->buddy_hash, 5, buddy_hash_compare, buddy_hash_cleanup);
	return (0);
}

void buddy_destroy(struct aim_priv *priv) {
	struct buddy_pref *pref = priv->buddy_pref;

	if (pref != NULL) {
		hash_destroy(&pref->buddy_hash);

		dlist_destroy(pref->group_list, NULL, group_cleanup);
		dlist_destroy(pref->permit_list, NULL, generic_cleanup);
		dlist_destroy(pref->block_list, NULL, generic_cleanup);

		pref->group_list = NULL;
		pref->permit_list = NULL;
		pref->block_list = NULL;
		priv->buddy_pref = NULL;

		free(pref);
	}
}

void buddy_update(	struct aim_priv *priv,
					struct buddy *buddy,
					aim_userinfo_t *userinfo)
{
	aim_update_buddy(buddy, userinfo);
	blist_update_label(priv->blist, buddy->blist_line);
}

void buddy_online(	struct aim_priv *priv,
					struct buddy *buddy,
					aim_userinfo_t *userinfo)
{
	buddy->last_seen = time(NULL);
	buddy->group->num_online++;
	buddy_update(priv, buddy, userinfo);
	blist_add(priv->blist, buddy);
}

void buddy_offline(struct aim_priv *priv, struct buddy *buddy) {
	buddy->last_seen = time(NULL);
	buddy->status = STATUS_OFFLINE;
	buddy->signon_time = 0;
	buddy->idle_time = 0;
	buddy->warn_level = 0;
	buddy->group->num_online--;

	blist_del(priv->blist, buddy);
}

void buddy_offline_all(struct aim_priv *priv) {
	struct buddy_pref *pref = priv->buddy_pref;
	dlist_t *gcur;

	if (pref == NULL)
		return;

	gcur = pref->group_list;
	while (gcur != NULL) {
		struct bgroup *gr = gcur->data;
		dlist_t *bcur = gr->member_list;

		while (bcur != NULL) {
			struct buddy *buddy = bcur->data;

			if (buddy->status != STATUS_OFFLINE)
				buddy_offline(priv, bcur->data);

			bcur = bcur->next;
		}

		gcur = gcur->next;
	}
}

int buddy_add_block(struct aim_priv *priv, char *screen_name, int aim) {
	dlist_t *node;
	char nname[NUSER_LEN];
	struct buddy_pref *pref = priv->buddy_pref;

	if (pref == NULL)
		return (-1);

	normalize(screen_name, nname, sizeof(nname));

	node = dlist_find(pref->block_list, nname, buddy_rights_compare);
	if (node != NULL)
		return (-1);

	pref->block_list = dlist_add_tail(pref->block_list, xstrdup(nname));

	if (aim)
		aim_add_block(priv, nname);

	return (0);
}

int buddy_remove_block(struct aim_priv *priv, char *screen_name, int aim) {
	dlist_t *node;
	char nname[NUSER_LEN];
	struct buddy_pref *pref = priv->buddy_pref;

	if (pref == NULL)
		return (-1);

	normalize(screen_name, nname, sizeof(nname));

	node = dlist_find(pref->block_list, nname, buddy_rights_compare);
	if (node == NULL)
		return (-1);

	if (aim)
		aim_remove_block(priv, nname);

	free(node->data);
	pref->block_list = dlist_remove(pref->block_list, node);
	return (0);
}

int buddy_add_permit(struct aim_priv *priv, char *screen_name, int aim) {
	dlist_t *node;
	char nname[NUSER_LEN];
	struct buddy_pref *pref = priv->buddy_pref;

	if (pref == NULL)
		return (-1);

	normalize(screen_name, nname, sizeof(nname));

	node = dlist_find(pref->permit_list, nname, buddy_rights_compare);
	if (node != NULL)
		return (-1);

	if (aim)
		aim_add_permit(priv, nname);

	pref->permit_list = dlist_add_tail(pref->permit_list, xstrdup(nname));
	return (0);
}

int buddy_remove_permit(struct aim_priv *priv, char *screen_name, int aim) {
	dlist_t *node;
	char nname[NUSER_LEN];
	struct buddy_pref *pref = priv->buddy_pref;

	if (pref == NULL)
		return (-1);

	normalize(screen_name, nname, sizeof(nname));

	node = dlist_find(pref->permit_list, nname, buddy_rights_compare);
	if (node == NULL)
		return (-1);

	if (aim)
		aim_remove_permit(priv, nname);

	free(node->data);
	pref->permit_list = dlist_remove(pref->permit_list, node);
	return (0);
}

void buddy_update_idle(struct aim_priv *priv) {
	struct buddy_pref *pref = priv->buddy_pref;
	dlist_t *gcur;

	if (pref == NULL)
		return;

	gcur = pref->group_list;
	while (gcur != NULL) {
		struct bgroup *gr = gcur->data;
		dlist_t *bcur = gr->member_list;

		while (bcur != NULL) {
			struct buddy *buddy = bcur->data;

			if (buddy->idle_time > 0) {
				buddy->idle_time++;
				blist_update_label(priv->blist, buddy->blist_line);
			}

			bcur = bcur->next;
		}

		gcur = gcur->next;
	}
}
